"""Keeps track of whether species are making progress and helps remove ones that are not."""
import sys

from neat.config import ConfigParameter, DefaultClassConfig
from neat.math_util import stat_functions

# TODO: Add a method for the user to change the "is stagnant" computation.


class DefaultStagnation(DefaultClassConfig):
    """Keeps track of whether species are making progress and helps remove ones that are not."""
    @classmethod
    def parse_config(cls, param_dict):
        return DefaultClassConfig(param_dict,
                                  [ConfigParameter('species_fitness_func', str, 'mean'),
                                   ConfigParameter('max_stagnation', int, 15),
                                   ConfigParameter('species_elitism', int, 0)])

    def __init__(self, config, reporters):
        # pylint: disable=super-init-not-called
        self.stagnation_config = config

        self.species_fitness_func = stat_functions.get(config.species_fitness_func)
        if self.species_fitness_func is None:
            raise RuntimeError(
                "Unexpected species fitness func: {0!r}".format(config.species_fitness_func))

        self.reporters = reporters

    def update(self, species_set, generation):
        """
        Required interface method. Updates species fitness history information,
        checking for ones that have not improved in max_stagnation generations,
        and - unless it would result in the number of species dropping below the configured
        species_elitism parameter if they were removed,
        in which case the highest-fitness species are spared -
        returns a list with stagnant species marked for removal.
        """
        species_data = []
        for sid, s in species_set.species.items():
            if s.fitness_history:
                prev_fitness = max(s.fitness_history)
            else:
                prev_fitness = -sys.float_info.max

            s.fitness = self.species_fitness_func(s.get_fitnesses())
            s.fitness_history.append(s.fitness)
            s.adjusted_fitness = None
            if prev_fitness is None or s.fitness > prev_fitness:
                s.last_improved = generation

            species_data.append((sid, s))

        # Sort in ascending fitness order.
        species_data.sort(key=lambda x: x[1].fitness)

        result = []
        species_fitnesses = []
        num_non_stagnant = len(species_data)
        for idx, (sid, s) in enumerate(species_data):
            # Override stagnant state if marking this species as stagnant would
            # result in the total number of species dropping below the limit.
            # Because species are in ascending fitness order, less fit species
            # will be marked as stagnant first.
            stagnant_time = generation - s.last_improved
            is_stagnant = False
            if num_non_stagnant > self.stagnation_config.species_elitism:
                is_stagnant = stagnant_time >= self.stagnation_config.max_stagnation

            if (len(species_data) - idx) <= self.stagnation_config.species_elitism:
                is_stagnant = False

            if is_stagnant:
                num_non_stagnant -= 1

            result.append((sid, s, is_stagnant))
            species_fitnesses.append(s.fitness)

        return result



# -*- coding: utf-8 -*-
# ************************************************************
# aitk.robots: Python robot simulator
#
# Copyright (c) 2021 AITK Developers
#
# https://github.com/ArtificialIntelligenceToolkit/aitk.robots
# ************************************************************

import math

from ..utils import (
    distance,
    display,
    PI_OVER_180,
    PI_OVER_2,
    TWO_PI,
    ONE80_OVER_PI,
    rotate_around,
    uniform_angle,
)
from .base import BaseDevice

from aitk.utils import Color

WHITE = Color("white")

class Camera(BaseDevice):
    def __init__(
        self,
        width=64,
        height=32,
        fov=30,
        colorsFadeWithDistance=1.0,
        sizeFadeWithDistance=0.8,
        reflectGround=True,
        reflectSky=False,
        max_range=1000,
        name="camera",
        samples=1,
        **kwargs
    ):
        """
        A camera device.

        Args:
            width (int): width of camera in pixels
            height (int): height of camera in pixels
            fov (float): width of camera field of view in degrees. Can be
                180 or even 360 for wide angle cameras.
            colorsFadeWithDistance (float): colors get darker
                faster with larger value
            sizeFadeWithDistance (float): size gets smaller faster
                with with larger value
            reflectGround (bool): ground reflects for 3D point cloud
            reflectSky (bool): sky reflects for 3D point cloud
            max_range (int): maximum range of camera
            name (str): the name of the camera
            samples (int): how many pixels should it sample

        Note: currently the camera faces forward.
        """
        config = {
            "width": width,
            "height": height,
            "fov": fov,
            "colorsFadeWithDistance": colorsFadeWithDistance,
            "sizeFadeWithDistance": sizeFadeWithDistance,
            "reflectGround": reflectGround,
            "reflectSky": reflectSky,
            "max_range": max_range,
            "name": name,
            "samples": samples,
        }
        config.update(kwargs)
        self._watcher = None
        self.robot = None
        self.initialize()
        self.from_json(config)

    def initialize(self):
        """
        Internal method to set all settings to default values.
        """
        # FIXME: camera is fixed at (0,0) facing forward
        self.type = "camera"
        self.bulbs_are_visible = True
        self.food_is_visible = True
        self.time = 0.0
        self.cameraShape = [64, 32]
        self.position = [0, 0]
        self.max_range = 1000
        self.samples = 1
        self.name = "camera"
        # 0 = no fade, 1.0 = max fade
        self.colorsFadeWithDistance = 0.9
        self.sizeFadeWithDistance = 1.0
        self.reflectGround = True
        self.reflectSky = False
        self.set_fov(60)  # degrees
        self.reset()

    def reset(self):
        """
        Internal method to reset the camera data.
        """
        self.hits = [[] for i in range(self.cameraShape[0])]

    def from_json(self, config):
        """
        Set the settings from a device config.

        Args:
            config (dict): a config dictionary
        """
        valid_keys = set([
            "width", "name", "height", "colorsFadeWithDistance",
            "sizeFadeWithDistance", "reflectGround", "reflectSky",
            "fov", "max_range", "samples", "position", "class"
        ])
        self.verify_config(valid_keys, config)

        if "width" in config:
            self.cameraShape[0] = config["width"]
        if "height" in config:
            self.cameraShape[1] = config["height"]

        if "colorsFadeWithDistance" in config:
            self.colorsFadeWithDistance = config["colorsFadeWithDistance"]
        if "sizeFadeWithDistance" in config:
            self.sizeFadeWithDistance = config["sizeFadeWithDistance"]
        if "reflectGround" in config:
            self.reflectGround = config["reflectGround"]
        if "reflectSky" in config:
            self.reflectSky = config["reflectSky"]
        if "fov" in config:
            self.set_fov(config["fov"])  # degrees
        if "max_range" in config:
            self.max_range = config["max_range"]
        if "samples" in config:
            self.samples = config["samples"]
        if "name" in config:
            self.name = config["name"]
        if "position" in config:
            self.position = config["position"]

    def to_json(self):
        """
        Save the internal settings to a config dictionary.
        """
        return {
            "class": self.__class__.__name__,
            "width": self.cameraShape[0],
            "height": self.cameraShape[1],
            "colorsFadeWithDistance": self.colorsFadeWithDistance,
            "sizeFadeWithDistance": self.sizeFadeWithDistance,
            "reflectGround": self.reflectGround,
            "reflectSky": self.reflectSky,
            "fov": self.fov * ONE80_OVER_PI,  # save in degrees
            "max_range": self.max_range,
            "samples": self.samples,
            "name": self.name,
            "position": self.position,
        }

    def __repr__(self):
        return "<Camera %r size=(%r,%r), fov=%r>" % (
            self.name,
            self.cameraShape[0],
            self.cameraShape[1],
            round(self.fov * ONE80_OVER_PI, 2),
        )

    def watch(self, **kwargs):
        """
        Create a dynamically updating view
        of this device.

        Args:
            title (str): title of device
        """
        if self.robot is None or self.robot.world is None:
            raise Exception("can't watch device until added to robot, and robot is in world")

        # Make the watcher if doesn't exist:
        self.get_widget(**kwargs)
        self._watcher.watch()

    def get_widget(self, **kwargs):
        """
        Return the dynamically updating widget.

        Args:
            title (str): title of device
        """
        from ..watchers import CameraWatcher

        if self._watcher is None:
            self._watcher = CameraWatcher(self, **kwargs)
            self.robot.world._watchers.append(self._watcher)
            return self._watcher.get_widget()
        else:
            return self._watcher.get_widget(**kwargs)

    def _step(self, time_step):
        pass

    def _get_visible_area(self):
        """
        What are the ranges of the field of view?
        Return a list of (p1, p2) that represent lines
        across the background, from front to back,
        down the edges of the field of view.
        """
        step = 1
        all_points = []
        for angle in [self.fov / 2, -self.fov / 2]:
            points = []
            dx, dy = rotate_around(0, 0, step, self.robot.a + angle)
            cx, cy = self.robot.x, self.robot.y
            x, y = cx, cy
            for i in range(0, self.max_range, step):
                points.append((x, y))
                x += dx
                y += dy
            all_points.append(points)
        return zip(*all_points)

    def update(self, draw_list=None):
        """
        Cameras operate in a lazy way: they don't actually update
        until needed because they are so expensive.

        Args:
            draw_list (list): optional. If given, then the
                method can add to it for drawing later.
        """
        pass

    def _update(self):
        # Update timestamp:
        self.time = self.robot.world.time

        for i in range(self.cameraShape[0]):
            angle = i / self.cameraShape[0] * self.fov - self.fov / 2
            self.hits[i] = self.robot.cast_ray(
                self.robot.x,
                self.robot.y,
                PI_OVER_2 - self.robot.a - angle,
                self.max_range,
            )


    def draw(self, backend):
        """
        Draw the device on the backend.

        Args:
            backend (Backend): an aitk drawing backend

        NOTE: Currently, cameras are fixed at 0,0 and face forwards.
        """
        backend.set_fill(Color(0, 64, 0))
        backend.strokeStyle(None, 0)
        backend.draw_rect(5.0, -3.33, 1.33, 6.33)

        # Note angle in sim world is opposite in graphics:
        hits = self.robot.cast_ray(
            self.robot.x,
            self.robot.y,
            PI_OVER_2 - self.robot.a + self.fov / 2,
            self.max_range,
        )
        if hits:
            p = rotate_around(0, 0, hits[-1].distance, -self.fov / 2,)
        else:
            p = rotate_around(0, 0, self.max_range, -self.fov / 2,)
        backend.draw_line(0, 0, p[0], p[1])

        # Note angle in sim world is opposite in graphics:
        hits = self.robot.cast_ray(
            self.robot.x,
            self.robot.y,
            PI_OVER_2 - self.robot.a - self.fov / 2,
            self.max_range,
        )
        if hits:
            p = rotate_around(0, 0, hits[-1].distance, self.fov / 2,)
        else:
            p = rotate_around(0, 0, self.max_range, self.fov / 2,)
        backend.draw_line(0, 0, p[0], p[1])

    def _find_closest_wall(self, hits):
        for hit in reversed(hits):  # reverse make it closest first
            if hit.height < 1.0:  # skip non-walls
                continue
            return hit.distance
        return float("inf")

    def _get_ground_color(self, area, i, j):
        if self.robot.world._ground_image is not None and area is not None:
            # i is width ray (camera width),
            # j is distance (height of camera/2, 64 to 128)
            dist = round(
                ((self.cameraShape[1] - j) / self.cameraShape[1] / 3) * len(area)
            )
            visible_width_points = area[dist]
            p1, p2 = visible_width_points
            # get a position i/width on line
            # minx, maxx = sorted([p1[0], p2[0]])
            # miny, maxy = sorted([p1[1], p2[1]])
            spanx = abs(p1[0] - p2[0])
            spany = abs(p1[1] - p2[1])
            if p1[0] < p2[0]:
                x = round(
                    (p1[0] + spanx * (1.0 - i / self.cameraShape[0]))
                    * self.robot.world.scale
                )
            else:
                x = round(
                    (p1[0] - spanx * (1.0 - i / self.cameraShape[0]))
                    * self.robot.world.scale
                )
            if p1[1] < p2[1]:
                y = round(
                    (p1[1] + spany * (1.0 - i / self.cameraShape[0]))
                    * self.robot.world.scale
                )
            else:
                y = round(
                    (p1[1] - spany * (1.0 - i / self.cameraShape[0]))
                    * self.robot.world.scale
                )
            # find that pixel
            if (0 <= x < (self.robot.world.width - 1) * self.robot.world.scale) and (
                0 <= y < (self.robot.world.height - 1) * self.robot.world.scale
            ):
                # c = Color(*self.robot.world._ground_image_pixels[(x, y)])
                # self.robot.world._ground_image_pixels[(x, y)] = (0, 0, 0)
                # return c

                sum = Color(0)
                count = 0
                # Need more sampling as distance increases:
                for j in range(self.samples):
                    try:
                        c = Color(*self.robot.world._ground_image_pixels[(x + j, y)])
                        # self.robot.world._ground_image_pixels[(x + j, y)] = (0, 0, 0)
                        count += 1
                    except Exception:
                        continue
                    sum += c
                return sum / count

        return self.robot.world.ground_color

    def display(self, type="color"):
        """
        Display an image of the camera.

        Args:
            type (str): the return type of image. Can be "color"
                or "depth"
        """
        image = self.get_image(type=type)
        display(image)

    def get_image(self, type="color"):
        """
        Get an image of the camera.

        Args:
            type (str): the return type of image. Can be "color"
                or "depth"
        """
        try:
            from PIL import Image
        except ImportError:
            print("Pillow (PIL) module not available; get_image() unavailable")
            return

        # Lazy; only get the data when we need it:
        self._update()
        if self.robot.world._ground_image is not None:
            area = list(self._get_visible_area())
        else:
            area = None
        pic = Image.new("RGBA", (self.cameraShape[0], self.cameraShape[1]))
        pic_pixels = pic.load()
        # FIXME: probably should have a specific size rather than scale it to world
        size = max(self.robot.world.width, self.robot.world.height)
        hcolor = None
        # draw non-robot walls first:
        for i in range(self.cameraShape[0]):
            hits = [hit for hit in self.hits[i] if hit.height == 1.0]  # only walls
            if len(hits) == 0:
                continue
            hit = hits[-1]  # get closest
            high = None
            hcolor = None
            if hit:
                if self.fov < PI_OVER_2:
                    # Orthoginal distance to camera:
                    angle = hit.angle
                    hit_distance = abs(hit.distance * math.sin(angle))
                else:
                    hit_distance = hit.distance

                distance_ratio = 1.0 - hit_distance / size
                s = distance_ratio * self.sizeFadeWithDistance
                sc = distance_ratio * self.colorsFadeWithDistance
                if type == "color":
                    r = hit.color.red * sc
                    g = hit.color.green * sc
                    b = hit.color.blue * sc
                elif type == "depth":
                    r = 255 * distance_ratio
                    g = 255 * distance_ratio
                    b = 255 * distance_ratio
                else:
                    avg = (hit.color.red + hit.color.green + hit.color.blue) / 3.0
                    r = avg * sc
                    g = avg * sc
                    b = avg * sc
                hcolor = Color(r, g, b)
                high = (1.0 - s) * self.cameraShape[1]
            else:
                high = 0

            horizon = self.cameraShape[1] / 2
            for j in range(self.cameraShape[1]):
                dist = max(min(abs(j - horizon) / horizon, 1.0), 0.0)
                if j < high / 2:  # sky
                    if type == "depth":
                        if self.reflectSky:
                            color = Color(255 * dist)
                        else:
                            color = Color(0)
                    elif type == "color":
                        color = Color(0, 0, 128)
                    else:
                        color = Color(128 / 3)
                    pic_pixels[i, j] = color.to_tuple()
                elif j < self.cameraShape[1] - high / 2:  # hit
                    if hcolor is not None:
                        pic_pixels[i, j] = hcolor.to_tuple()
                else:  # ground
                    if type == "depth":
                        if self.reflectGround:
                            color = Color(255 * dist)
                        else:
                            color = Color(0)
                    elif type == "color":
                        color = self._get_ground_color(area, i, j)
                    else:
                        color = Color(128 / 3)
                    pic_pixels[i, j] = color.to_tuple()

        # Other robots, draw on top of walls:
        self.obstacles = {}
        for i in range(self.cameraShape[0]):
            closest_wall_dist = self._find_closest_wall(self.hits[i])
            hits = [hit for hit in self.hits[i] if hit.height < 1.0]  # obstacles
            for hit in hits:
                if hit.distance > closest_wall_dist:
                    # Behind wall, try next
                    continue

                if self.fov < PI_OVER_2:
                    angle = hit.angle
                    hit_distance = abs(hit.distance * math.sin(angle))
                else: # perspective
                    hit_distance = hit.distance

                distance_ratio = 4.0 - hit_distance / size
                s = distance_ratio * self.sizeFadeWithDistance
                sc = distance_ratio * self.colorsFadeWithDistance
                distance_to = self.cameraShape[1] / 2 * (4.0 - sc)
                # scribbler was 30, so 0.23 height ratio
                # height is ratio, 0 to 1
                height = round(hit.height * self.cameraShape[1] / 2.0 * s)
                if type == "color":
                    r = hit.color.red * sc
                    g = hit.color.green * sc
                    b = hit.color.blue * sc
                elif type == "depth":
                    r = 255 * distance_ratio
                    g = 255 * distance_ratio
                    b = 255 * distance_ratio
                else:
                    avg = (hit.color.red + hit.color.green + hit.color.blue) / 3.0
                    r = avg * sc
                    g = avg * sc
                    b = avg * sc
                hcolor = Color(r, g, b)
                horizon = self.cameraShape[1] / 2
                self._record_obstacle(
                    hit.robot,
                    i,
                    self.cameraShape[1] - 1 - round(distance_to),
                    self.cameraShape[1] - height - 1 - 1 - round(distance_to),
                )
                if not hit.robot.has_image():
                    for j in range(height):
                        pic_pixels[
                            i, self.cameraShape[1] - j - 1 - round(distance_to)
                        ] = hcolor.to_tuple()
        if self.bulbs_are_visible:
            self._show_bulbs(pic)
        if self.food_is_visible:
            self._show_food(pic)
        self._show_obstacles(pic)
        return pic

    def _show_bulbs(self, image):
        from PIL import ImageDraw, Image

        draw_list = []
        for bulb in self.robot.world._get_light_sources(all=True):
            if bulb.robot is self.robot:
                # You can't sense your own bulbs
                continue

            # cast a ray, see if it hits this robot
            # if so, we can see it
            # ignore boundary boxes around robots that contain the bulb
            bulb_x, bulb_y = bulb.get_position()
            hits = self.robot.cast_ray(bulb_x, bulb_y, 0, self.max_range,
                                       self.robot.x, self.robot.y, ignore_robots=[bulb.robot,
                                                                                  self.robot])
            if len(hits) == 0:
                draw_list.append((bulb, bulb_x, bulb_y))

        if len(draw_list) > 0:
            layer = Image.new('RGBA', self.cameraShape, (0, 0, 0, 0))
            drawing = ImageDraw.Draw(layer, "RGBA")
            size = max(self.robot.world.width, self.robot.world.height)
            color = Color(bulb.color)
            #color.alpha = 128
            for (bulb, bulb_x, bulb_y) in draw_list:
                angle = uniform_angle(math.pi * 3/2 - math.atan2(self.robot.x - bulb_x,
                                                                 self.robot.y - bulb_y))
                min_angle = uniform_angle(self.robot.a - self.fov/2)
                max_angle = uniform_angle(self.robot.a + self.fov/2)
                if max_angle < min_angle:
                    if 0 <= angle < max_angle:
                        angle += TWO_PI
                    max_angle += TWO_PI

                if min_angle < angle < max_angle:
                    span = max_angle - min_angle
                    x = int((angle - min_angle)/span * self.cameraShape[0])
                    hit_distance = distance(self.robot.x, self.robot.y, bulb_x, bulb_y)
                    distance_ratio = 1.0 - hit_distance / size
                    s = distance_ratio * self.sizeFadeWithDistance
                    high = (1.0 - s) * self.cameraShape[1]
                    y = (self.cameraShape[1] / 2) + high / 2
                    radius = 5
                    drawing.ellipse((x - radius, y - radius,
                                     x + radius, y + radius),
                                    fill=color.to_tuple())
            image.paste(Image.alpha_composite(image, layer))

    def _show_food(self, image):
        from PIL import ImageDraw, Image

        draw_list = []
        for food in self.robot.world._food:
            hits = self.robot.cast_ray(food.x, food.y, 0, self.max_range,
                                       self.robot.x, self.robot.y)
            if len(hits) == 0:
                draw_list.append((food.x, food.y, food.standard_deviation))

        if len(draw_list) > 0:
            layer = Image.new('RGBA', self.cameraShape, (0, 0, 0, 0))
            drawing = ImageDraw.Draw(layer, "RGBA")
            size = max(self.robot.world.width, self.robot.world.height)
            color = Color(WHITE)
            #color.alpha = 128
            for (x, y, sd) in draw_list:
                angle = uniform_angle(math.pi * 3/2 - math.atan2(self.robot.x - x,
                                                                 self.robot.y - y))
                min_angle = uniform_angle(self.robot.a - self.fov/2)
                max_angle = uniform_angle(self.robot.a + self.fov/2)
                if max_angle < min_angle:
                    if 0 <= angle < max_angle:
                        angle += TWO_PI
                    max_angle += TWO_PI

                if min_angle < angle < max_angle:
                    span = max_angle - min_angle
                    x = int((angle - min_angle)/span * self.cameraShape[0])
                    hit_distance = distance(self.robot.x, self.robot.y, x, y)
                    distance_ratio = 1.0 - hit_distance / size
                    s = distance_ratio * self.sizeFadeWithDistance
                    high = (1.0 - s) * self.cameraShape[1]
                    y = (self.cameraShape[1] / 2) + high / 2
                    radius = 5
                    drawing.rectangle((x - radius, y - radius,
                                       x + radius, y + radius),
                                      fill=color.to_tuple())
            image.paste(Image.alpha_composite(image, layer))

    def _show_obstacles(self, image):
        # FIXME: show back to front
        # FIXME: how to show when partially behind wall?
        # data = {"robot": obj.has_image(), "min_x", "min_y", "max_x", "max_y"}
        for data in self.obstacles.values():
            if data["robot"].has_image():
                # the angle to me + offset for graphics + the robot angle:
                radians = (
                    math.atan2(
                        data["robot"].x - self.robot.x, data["robot"].y - self.robot.y
                    )
                    + PI_OVER_2
                    + data["robot"].a
                )
                degrees = round(radians * ONE80_OVER_PI)
                picture = data["robot"].get_image_3d(degrees)  # degrees
                x1, y1 = data["min_x"], data["min_y"]  # noqa: F841
                x2, y2 = data["max_x"], data["max_y"]
                try:  # like too small
                    picture.thumbnail((x2 - x1, 10000))  # to keep aspect ratio
                    # picture = picture.resize((x2 - x1, y2 - y1))
                    x3 = x2 - picture.height
                    y3 = y2 - picture.width
                    image.paste(picture, (x3, y3), picture)
                except Exception:
                    print("Exception in processing image")

    def _record_obstacle(self, robot, x, y1, y2):
        if robot.name not in self.obstacles:
            self.obstacles[robot.name] = {
                "robot": robot,
                "max_x": float("-inf"),
                "max_y": float("-inf"),
                "min_x": float("inf"),
                "min_y": float("inf"),
            }
        self.obstacles[robot.name]["max_x"] = max(
            self.obstacles[robot.name]["max_x"], x
        )
        self.obstacles[robot.name]["min_x"] = min(
            self.obstacles[robot.name]["min_x"], x
        )
        self.obstacles[robot.name]["max_y"] = max(
            self.obstacles[robot.name]["max_y"], y1, y2
        )
        self.obstacles[robot.name]["min_y"] = min(
            self.obstacles[robot.name]["min_y"], y1, y2
        )

    def get_point_cloud(self):
        """
        Get a 3D point cloud from the camera data.

        Returns a list of [x, y, distance, red, green, blue]
        for each (x,y) of the camera.
        """
        depth_pic = self.get_image("depth")
        depth_pixels = depth_pic.load()
        color_pic = self.get_image("color")
        color_pixels = color_pic.load()
        points = []
        for x in range(self.cameraShape[0]):
            for y in range(self.cameraShape[1]):
                dist_color = depth_pixels[x, y]
                color = color_pixels[x, y]
                if dist_color[0] != 255:
                    points.append(
                        [
                            self.cameraShape[0] - x - 1,
                            self.cameraShape[1] - y - 1,
                            dist_color[0],
                            color[0],
                            color[1],
                            color[2],
                        ]
                    )
        return points

    def get_fov(self):
        """
        Get the field of view angle in degrees.
        """
        return self.fov * ONE80_OVER_PI

    def set_fov(self, angle):
        """
        Set the field of view angle in degrees of the camera.

        Args:
            angle (float): angle in degrees of field of view
        """
        # given in degrees
        # save in radians
        # scale = min(max(angle / 6.0, 0.0), 1.0)
        self.fov = angle * PI_OVER_180
        # self.sizeFadeWithDistance = scale
        self.reset()

    def set_size(self, width, height):
        """
        Set the height and width of the camera in pixels.

        Args:
            width (int): width of camera in pixels
            height (int): height of camera in pixels
        """
        self.cameraShape[0] = width
        self.cameraShape[1] = height
        self.reset()

    def set_max(self, max_range):
        """
        Set the maximum distance the camera can see.

        Args:
            max_range (float): distance (in CM) the camera can see
        """
        self.max_range = max_range

    def set_width(self, width):
        """
        Set the width of the camera in pixels.

        Args:
            width (int): width of camera in pixels
        """
        self.cameraShape[0] = width
        self.reset()

    def set_height(self, height):
        """
        Set the height of the camera in pixels.

        Args:
            height (int): height of camera in pixels
        """
        self.cameraShape[1] = height
        self.reset()

    def set_name(self, name):
        """
        Set the name of the camera.

        Args:
            name (str): the name of the camera
        """
        self.name = name

    def get_name(self):
        """
        Get the name of the camera.
        """
        return self.name

    def get_width(self):
        """
        Get the width in pixels of the camera.
        """
        return self.cameraShape[0]

    def get_height(self):
        """
        Get the height in pixels of the camera.
        """
        return self.cameraShape[1]

    def get_max(self):
        """
        Get the maximum distance in CM the camera can see.
        """
        return self.max_range


class GroundCamera(Camera):
    def __init__(self, width=15, height=15, name="ground-camera", **kwargs):
        """
        A downward-facing camera device.

        Args:
            width (int): width of camera in pixels
            height (int): height of camera in pixels
            name (str): the name of the camera
        """
        config = {
            "width": width,
            "height": height,
            "name": name,
        }
        self._watcher = None
        self.robot = None
        self.initialize()
        self.from_json(config)

    def initialize(self):
        """
        Internal method to set all settings to default values.
        """
        self.type = "ground-camera"
        self.time = 0.0
        self.cameraShape = [15, 15]
        self.name = "ground-camera"

    def from_json(self, config):
        """
        Set the settings from a device config.

        Args:
            config (dict): a config dictionary
        """
        if "width" in config:
            self.cameraShape[0] = config["width"]
        if "height" in config:
            self.cameraShape[1] = config["height"]
        if "name" in config:
            self.name = config["name"]

    def to_json(self):
        """
        Save the internal settings to a config dictionary.
        """
        return {
            "class": self.__class__.__name__,
            "width": self.cameraShape[0],
            "height": self.cameraShape[1],
            "name": self.name,
        }

    def __repr__(self):
        return "<GroundCamera %r size=(%r,%r)>" % (
            self.name,
            self.cameraShape[0],
            self.cameraShape[1],
        )

    def update(self, draw_list=None):
        """
        Update the device.

        Args:
            draw_list (list): optional. If given, then the
                method can add to it for drawing later.

        Cameras operate in a lazy way: they don't actually update
        until needed because they are so expensive.
        """
        pass

    def draw(self, backend):
        """
        Draw the device on the backend.

        Args:
            backend (Backend): an aitk drawing backend
        """
        left = -(self.cameraShape[0] // 2) / self.robot.world.scale
        upper = -(self.cameraShape[1] // 2) / self.robot.world.scale
        right = (self.cameraShape[0] // 2) / self.robot.world.scale
        lower = (self.cameraShape[1] // 2) / self.robot.world.scale
        backend.strokeStyle(Color(128), 1)
        backend.draw_line(left, upper, right, upper)
        backend.draw_line(right, upper, right, lower)
        backend.draw_line(right, lower, left, lower)
        backend.draw_line(left, lower, left, upper)

    def get_image(self, type=None):
        """
        Get an image of the camera.

        Args:
            type (str): the return type of image. Can be "color"
                or "depth"
        """
        # FIXME: would be faster to trim image down
        # before rotating
        center = (
            self.robot.x * self.robot.world.scale,
            self.robot.y * self.robot.world.scale,
        )
        rotated_image = self.robot.world._ground_image.rotate(
            (self.robot.a - math.pi / 4 * 6) * (ONE80_OVER_PI), center=center,
        )
        left = center[0] - self.cameraShape[0] // 2
        right = center[0] + self.cameraShape[0] // 2
        upper = center[1] - self.cameraShape[1] // 2
        lower = center[1] + self.cameraShape[1] // 2
        return rotated_image.crop((left, upper, right, lower))



"""
Runs evaluation functions in parallel subprocesses
in order to evaluate multiple genomes at once.
"""
from multiprocessing import Pool


class ParallelEvaluator(object):
    def __init__(self, num_workers, eval_function, timeout=None):
        """
        eval_function should take one argument, a tuple of
        (genome object, config object), and return
        a single float (the genome's fitness).
        """
        self.num_workers = num_workers
        self.eval_function = eval_function
        self.timeout = timeout
        self.pool = Pool(num_workers)

    def __del__(self):
        self.pool.close() # should this be terminate?
        self.pool.join()

    def evaluate(self, genomes, config):
        jobs = []
        for ignored_genome_id, genome in genomes:
            jobs.append(self.pool.apply_async(self.eval_function, (genome, config)))

        # assign the fitness back to each genome
        for job, (ignored_genome_id, genome) in zip(jobs, genomes):
            genome.fitness = job.get(timeout=self.timeout)



# -*- coding: utf-8 -*-
# ************************************************************
# aitk.robots: Python robot simulator
#
# Copyright (c) 2021 AITK Developers
#
# https://github.com/ArtificialIntelligenceToolkit/aitk.robots
# ************************************************************

import io
import math

from PIL import Image, ImageDraw, ImageFont

from aitk.utils import get_font, Color
from ..utils import arange, distance
from .base import Backend

BLACK = Color("black")
WHITE = Color("white")

DEFAULT_FONT_NAMES = (
    get_font("FreeMonoBold.ttf"),
    "arial.ttf",
    "Arial.ttf",
    "NotoSans-Regular.ttf",
    "/usr/share/fonts/truetype/freefont/FreeMonoBold.ttf",
    "/System/Library/Fonts/SFNSDisplay.ttf",
    "/Library/Fonts/Arial.ttf",
)


class PILBackend(Backend):
    # Specific to this class:

    def __init__(self, *args, **kwargs):
        self.widget = None
        super().__init__(*args, **kwargs)

    def initialize(self, **kwargs):
        self.matrix = []
        self.kwargs = kwargs
        self.font = None
        self.font_size = kwargs.get("font_size", int(12 * self._scale))
        for font_string_name in DEFAULT_FONT_NAMES:
            try:
                self.font = ImageFont.truetype(font_string_name, self.font_size)
                break
            except OSError:
                continue

        self.mode = kwargs.get("mode", "RGBA")  # "RGBA" or "RGB"
        self.format = kwargs.get(
            "format", "jpeg"
        )  # "png", "gif", or "jpeg" # png,gif doesn't do opacity?!

        if self.mode == "RGBA" and self.format == "jpeg":
            self.mode = "RGB"
            kwargs["mode"] = "RGB"

        self.status_height = 14
        self.image = Image.new(
            self.mode,
            size=(int(self.width * self._scale), int((self.height  + self.status_height) * self._scale)),
        )
        self.draw = ImageDraw.Draw(self.image, "RGBA")
        if self.font:
            (left, top, right, bottom) = self.draw.textbbox((0, 0), "0", font=self.font, anchor="ls")
            text_width = right - left
            text_height = top - bottom
            #text_width, text_height = self.draw.textsize("0", self.font)
            self.char_width = text_width / self._scale
            self.char_height = text_height / self._scale
        else:
            self.char_width = 5.8
            self.char_height = 10

    def update_dimensions(self, width, height, scale):
        if width != self.width or height != self.height or self._scale != scale:
            self.width = width
            self.height = height
            self._scale = scale
            self.initialize(**self.kwargs)

    # Canvas API:

    def draw_status(self, text):
        self.set_fill(BLACK)
        self.draw_rect(
            0,
            self.height,
            self.width,
            self.status_height,
        )
        self.set_fill(WHITE)
        self.text(text, 1, self.height - 1)

    def to_image(self, format=None):
        format = format if format is not None else self.format
        fp = io.BytesIO()
        self.image.save(fp, format=format)
        return fp.getvalue()

    def get_widget(self, width=None, height=None):
        from ipywidgets import Image

        if self.widget is None:
            self.widget = Image(value=self.to_image())
            self.widget.layout.margin = "auto"
            self.widget.layout.border = "10px solid rgb(0 177 255)"

        if width is not None:
            self.widget.layout.width = width
        if height is not None:
            self.widget.layout.height = height

        return self.widget

    def draw_watcher(self):
        if self.widget:
            self.widget.value = self.to_image()

    def flush(self):
        pass

    def get_image(self, time):
        return self.image

    # High-level API:

    def draw_image(self, image, x, y):
        self.image.paste(image, (x, y))

    def get_color(self, color):
        if isinstance(color, Color):
            return color.to_tuple()
        elif color != "":
            return color
        else:
            return None

    def get_line_width(self):
        return round(self.line_width * self._scale)

    def get_style(self, style):
        if style == "fill":
            return self.get_color(self.fill_style)
        elif style == "stroke":
            return self.get_color(self.stroke_style)

    def p(self, x, y):
        # Transform a point
        for matrix in self.matrix:
            for transform in reversed(matrix):
                if transform[0] == "translate":
                    x += transform[1]
                    y += transform[2]
                elif transform[0] == "rotate":
                    dist = distance(0, 0, x, y)
                    angle2 = math.atan2(-x, y)
                    angle = transform[1]
                    x = dist * math.cos(angle2 + angle + math.pi / 2)
                    y = dist * math.sin(angle2 + angle + math.pi / 2)
        return x * self._scale, y * self._scale

    def r(self, angle):
        # Transform an angle
        for matrix in self.matrix:
            for transform in reversed(matrix):
                if transform[0] == "translate":
                    pass
                elif transform[0] == "rotate":
                    angle += transform[1]
        return angle

    def draw_lines(self, points, stroke_style=None):
        self.stroke_style = stroke_style
        for i in range(len(points)):
            if i < len(points) - 2:
                self.draw_line(
                    points[i][0], points[i][1], points[i + 1][0], points[i + 1][1]
                )

    def draw_line(self, x1, y1, x2, y2):
        p1x, p1y = self.p(x1, y1)
        p2x, p2y = self.p(x2, y2)
        self.draw.line(
            (p1x, p1y, p2x, p2y),
            fill=self.get_style("stroke"),
            width=self.get_line_width(),
        )

    def clear(self):
        # self.fill_style = "white"
        self.draw_rect(0, 0, self.width, self.height)

    def text(self, t, x, y):
        x, y = self.p(x, y)
        self.draw.text((x, y), t, fill=self.get_style("fill"), font=self.font)

    def pushMatrix(self):
        self.matrix.append([])

    def popMatrix(self):
        self.matrix.pop()

    def scale(self, x, y):
        pass

    def resetScale(self):
        pass

    def draw_rect(self, x, y, width, height):
        p1x, p1y = self.p(x, y)
        p2x, p2y = self.p(x + width, y)
        p3x, p3y = self.p(x + width, y + height)
        p4x, p4y = self.p(x, y + height)

        self.draw.polygon(
            (p1x, p1y, p2x, p2y, p3x, p3y, p4x, p4y),
            fill=self.get_style("fill"),
            outline=self.get_style("stroke"),
        )

    def draw_ellipse(self, x, y, radiusX, radiusY):
        # Given as center and radius
        if radiusX == radiusY:
            x, y = self.p(x, y)

            p1x, p1y = (x - radiusX * self._scale, y - radiusY * self._scale)
            p2x, p2y = (x + radiusX * self._scale, y + radiusY * self._scale)

            minx = min(p1x, p2x)
            miny = min(p1y, p2y)
            maxx = max(p1x, p2x)
            maxy = max(p1y, p2y)

            self.draw.ellipse(
                (minx, miny, maxx, maxy),
                fill=self.get_style("fill"),
                outline=self.get_style("stroke"),
                width=self.get_line_width(),
            )
        else:
            self.draw_arc(x, y, radiusX, radiusY, 0, math.pi * 2, 12)

    def draw_arc(self, x, y, width, height, startAngle, endAngle, segments=5):
        points = [self.p(x, y)]

        for angle in arange(startAngle, endAngle, (endAngle - startAngle) / segments):
            point = (
                x + height * math.cos(angle),
                y + width * math.sin(angle),
            )
            points.append(self.p(point[0], point[1]))

        self.draw.polygon(
            points, fill=self.get_style("fill"),
        )

        self.draw.line(
            points[1:], fill=self.get_style("stroke"), width=self.get_line_width()
        )

    def beginShape(self):
        self.points = []

    def endShape(self):
        self.draw.polygon(
            self.points, fill=self.get_style("fill"), outline=self.get_style("stroke")
        )

    def vertex(self, x, y):
        self.points.append(self.p(x, y))

    def translate(self, x, y):
        self.matrix[-1].append(("translate", x, y))

    def rotate(self, angle):
        self.matrix[-1].append(("rotate", angle))



# -*- coding: utf-8 -*-
# ************************************************************
# aitk.robots: Python robot simulator
#
# Copyright (c) 2021 AITK Developers
#
# https://github.com/ArtificialIntelligenceToolkit/aitk.robots
# ************************************************************

from .base import Backend

# Implements all of the needed methods:


class DebugBackend(Backend):

    # Canvas API:
    def initialize(self, **kwargs):
        self.show_high = kwargs.get("show_high", True)
        self.show_low = kwargs.get("show_low", True)

    def flush(self):
        print("flush()")

    def get_image(self, time):
        print("get_image()")

    def watch(self):
        print("watch()")

    # High-level API (backend draw API)

    def draw_image(self, image, x, y):
        if self.show_high:
            print("draw_image%r" % ((image, x, y)))
        super().draw_image(image, x, y)

    def draw_lines(self, points, stroke_style=None):
        if self.show_high:
            print("draw_lines%r" % ((points, stroke_style),))
        super().draw_lines(points, stroke_style)

    def set_stroke_style(self, color):
        if self.show_high:
            print("set_stroke_style%r" % ((color),))
        super().set_stroke_style(color)

    def set_fill_style(self, color):
        if self.show_high:
            print("set_fill_style%r" % ((color),))
        super().set_fill_style(color)

    def clear(self):
        if self.show_high:
            print("clear()")
        super().clear()

    def set_font(self, style):
        if self.show_high:
            print("set_font%r" % ((style,),))
        super().set_font(style)

    def text(self, t, x, y):
        if self.show_high:
            print("text%r" % ((t, x, y),))
        super().text(t, x, y)

    def lineWidth(self, width):
        if self.show_high:
            print("lineWidth%r" % ((width),))
        super().lineWidth(width)

    def strokeStyle(self, color, width):
        if self.show_high:
            print("strokeStyle%r" % ((color, width),))
        super().strokeStyle(color, width)

    def make_stroke(self):
        if self.show_high:
            print("make_stroke()")
        super().make_stroke()

    def noStroke(self):
        if self.show_high:
            print("noStroke()")
        super().noStroke()

    def set_fill(self, color):
        if self.show_high:
            print("set_fill%r" % ((color),))
        super().set_fill(color)

    def noFill(self):
        if self.show_high:
            print("noFill()")
        super().noFill()

    def draw_line(self, x1, y1, x2, y2):
        if self.show_high:
            print("draw_line%r" % ((x1, y1, x2, y2),))
        super().draw_line(x1, y1, x2, y2)

    def pushMatrix(self):
        if self.show_high:
            print("pushMatrix()")
        super().pushMatrix()

    def popMatrix(self):
        if self.show_high:
            print("popMatrix()")
        super().popMatrix()

    def resetScale(self):
        if self.show_high:
            print("resetScale()")
        super().resetScale()

    def beginShape(self):
        if self.show_high:
            print("beginShape()")
        super().beginShape()

    def endShape(self):
        if self.show_high:
            print("endShape()")
        super().endShape()

    def vertex(self, x, y):
        if self.show_high:
            print("vertex%r" % ((x, y),))
        super().vertex(x, y)

    def draw_rect(self, x, y, width, height):
        if self.show_high:
            print("draw_rect%r" % ((x, y, width, height),))
        super().draw_rect(x, y, width, height)

    def draw_circle(self, x, y, radius):
        if self.show_high:
            print("draw_cirle%r" % ((x, y, radius),))
        super().draw_circle(x, y, radius)

    def draw_ellipse(self, x, y, radiusX, radiusY):
        if self.show_high:
            print("draw_ellipse%r" % ((x, y, radiusX, radiusY),))
        super().draw_ellipse(x, y, radiusX, radiusY)

    def draw_arc(self, x, y, width, height, startAngle, endAngle):
        if self.show_high:
            print("draw_arc%r" % ((x, y, width, height, startAngle, endAngle),))
        super().draw_arc(x, y, width, height, startAngle, endAngle)

    # Low-level API (HTML Canvas API):

    def arc(self, x, y, width, startAngle, endAngle):
        if self.show_low:
            print("    arc(%r,%r,%r,%r,%r)" % (x, y, width, startAngle, endAngle))

    def get_image_data(self):
        if self.show_low:
            print("    get_image_data()")

    def clear_rect(self, x, y, width, height):
        if self.show_low:
            print("    clear_rect(", x, y, width, height, ")")

    def fill_text(self, t, x, y):
        if self.show_low:
            print("    fill_text(", t, x, y, ")")

    def fill_rect(self, x, y, width, height):
        if self.show_low:
            print("    fill_rect(", x, y, width, height, ")")

    def fill(self):
        if self.show_low:
            print("    fill()")

    def stroke(self):
        if self.show_low:
            print("    stroke()")

    def move_to(self, x, y):
        if self.show_low:
            print("    move_to(", x, y, ")")

    def line_to(self, x, y):
        if self.show_low:
            print("    line_to(", x, y, ")")

    def save(self):
        if self.show_low:
            print("    save()")

    def restore(self):
        if self.show_low:
            print("    restore()")

    def translate(self, x, y):
        if self.show_low:
            print("    translate(", x, y, ")")

    def scale(self, xscale, yscale):
        if self.show_low:
            print("    scale(", xscale, yscale, ")")

    def set_transform(self, x, y, z, a, b, c):
        if self.show_low:
            print("    set_transform(", x, y, z, a, b, c, ")")

    def rotate(self, angle):
        if self.show_low:
            print("    rotate(", angle, ")")

    def begin_path(self):
        if self.show_low:
            print("    begin_path()")

    def close_path(self):
        if self.show_low:
            print("    close_path()")

    def ellipse(self, x, y, radiusX, radiusY, a, b, angle):
        if self.show_low:
            print("    ellipse(", x, y, radiusX, radiusY, a, b, angle, ")")

    def put_image_data(self, scaled, x, y):
        if self.show_low:
            print("    put_image_data(", scaled, x, y, ")")

    def create_image_data(self, width, height):
        if self.show_low:
            print("    create_image_data(", width, height, ")")



# -*- coding: utf-8 -*-
# ***********************************************************
# aitk.utils: Python utils for AI
#
# Copyright (c) 2021 AITK Developers
#
# https://github.com/ArtificialIntelligenceToolkit/aitk.utils
#
# ***********************************************************

import base64
import html
import io
import math
import os
import sys

try:
    from IPython.display import clear_output, display
except ImportError:
    clear_ouput = None


def array_to_image(array, colormap=None, channels="last", minmax=None):
    """
    Convert an array to a Python Image.

    Args:
        array (sequence): list, or array, of numbers
        colormap (str): optional, the name of the colormap to use
        channels (str): optional, "first" or "last"
        minmax (sequence of 2 numbers): optional, range of numbers
            to scale from, or None to compute dynamically
    """
    try:
        import numpy
        import PIL.Image
    except ImportError as exc:
        raise Exception(
            "The Python libraries PIL and numpy are required for converting an array into an image"
        ) from exc

    array = numpy.array(array)

    if channels == "first" and len(array.shape) == 3:
        array = numpy.moveaxis(array, 0, -1)

    if len(array.shape) == 3:
        if array.shape[-1] == 1:
            array = array.reshape((array.shape[0], array.shape[1]))
        elif array.shape[0] == 1:
            array = array.reshape((array.shape[1], array.shape[2]))
    elif len(array.shape) == 1:
        array = numpy.array([array])

    if minmax is None:
        minmax = array.min(), array.max()

    if colormap is not None:
        try:
            from matplotlib import cm
        except ImportError as exc:
            raise Exception(
                "The Python library matplotlib is required to use a colormap"
            ) from exc

        ## Need to be in range (0,1) for colormapping:
        if minmax[0] != minmax[1]:
            array.clip(*minmax)
        else:
            minmax = [minmax[0] - 1, minmax[1] + 1]
        array = rescale_array(array, minmax, (0, 1), "float")
        try:
            cm_hot = cm.get_cmap(colormap)
            array = cm_hot(array)
        except Exception:
            print("WARNING: invalid colormap; ignored")

        minmax = (0, 1)
        mode = "RGBA"
    else:
        mode = None

    array = rescale_array(array, minmax, (0, 255), "uint8")
    image = PIL.Image.fromarray(array, mode)
    return image


def rescale_array(array, old_range, new_range, dtype):
    """
    Given a numpy array in an old_range, rescale it
    into new_range, and make it an array of dtype.
    """
    try:
        import numpy
    except ImportError as exc:
        raise Exception(
            "The Python library numpy is required for rescaling array"
        ) from exc

    old_min, old_max = old_range
    if array.min() < old_min or array.max() > old_max:
        ## truncate:
        array = numpy.clip(array, old_min, old_max)
    new_min, new_max = new_range
    old_delta = float(old_max - old_min)
    new_delta = float(new_max - new_min)
    if old_delta == 0:
        return ((array - old_min) + (new_min + new_max) / 2).astype(dtype)
    else:
        return (new_min + (array - old_min) * new_delta / old_delta).astype(dtype)


def image_to_data(img_src, format="PNG"):
    # Convert to binary data:
    b = io.BytesIO()
    img_src.save(b, format=format)
    data = b.getvalue()
    data = base64.b64encode(data)
    if not isinstance(data, str):
        data = data.decode("latin1")
    return "data:image/%s;base64,%s" % (format, html.escape(data))


def gallery(
    images,
    labels="{index}",
    border_width=1,
    background_color=(255, 255, 255),
    return_type="display",
    clear=True,
    gallery_shape=None,
):
    """
    Construct a gallery (grid) of images. Can return an HTML table of images
    or a single Image.

    Args:
        images (sequence of Image or an array): sequence of PIL Images or arrays.
            If arrays, they will be processed by array_to_image()
        labels (str or sequence): optional, default "{index}"
        border_width (int): border around images
        background_color (list or tuple of 3 int): optional, represents RGB.
        return_type (str): "display" or None
        clear (bool): if format is None, then clear the display output
        gallery_shape (sequence of 2 int): optional, (cols, rows). One of
            row or col may be None, in order to set one of the dimensions.

    The labels argument can be a string containing the following special patterns
    that will be replaced:

    * "{index}" - the 0-based index of the item
    * "{count}" - index + 1
    * "{row}" - the row of the item
    * "{col}" - the column of the item

    If labels is not None, then gallery() will create a single image. If return_type
    is None then it will return the image rather than display it.

    If clear is True, then the output will be cleared before displaying.
    """
    try:
        import PIL.Image
        from IPython.display import HTML
    except ImportError:
        print("gallery() requires Pillow (Python Image Library) and IPython")
        return

    if len(images) == 0:
        return None

    if (gallery_shape is None) or (
        len(gallery_shape) == 2
        and (gallery_shape[0] is None)
        and (gallery_shape[1] is None)
    ):
        gallery_cols = math.ceil(math.sqrt(len(images)))
        gallery_rows = math.ceil(len(images) / gallery_cols)
    else:
        gallery_cols, gallery_rows = gallery_shape
        if gallery_cols is None:
            gallery_cols = math.ceil(len(images) / gallery_rows)
        elif gallery_rows is None:
            gallery_rows = math.ceil(len(images) / gallery_cols)

    # check that all images are images:
    _images = []
    for index in range(len(images)):
        image = images[index]
        if not isinstance(image, PIL.Image.Image):
            image = array_to_image(image)
        _images.append(image)
    images = _images

    if labels is None:
        size = images[0].size
        size = size[0] + (border_width * 2), size[1] + (border_width * 2)

        gallery_image = PIL.Image.new(
            mode="RGBA",
            size=(int(gallery_cols * size[0]), int(gallery_rows * size[1])),
            color=background_color,
        )

        for i, image in enumerate(images):
            if image.mode != "RGBA":
                image = image.convert("RGBA")
            location = (
                int((i % gallery_cols) * size[0]) + border_width,
                int((i // gallery_cols) * size[1]) + border_width,
            )
            gallery_image.paste(image, location)
        output = gallery_image
    else:
        if isinstance(labels, str):
            label_pattern = labels
            labels = [label_pattern for i in range(len(images))]

        table = "<table>"
        index = 0
        for row in range(gallery_rows):
            table += '<tr style="padding: %dpx">' % border_width
            for col in range(gallery_cols):
                if index < len(labels):
                    label = str(labels[index]).format(
                        **{"count": index + 1, "index": index, "row": row, "col": col}
                    )
                    table += '<td style="text-align: center; padding: %dpx">%s<br/>' % (
                        border_width,
                        label,
                    )
                    table += '<img src="%s" alt="%s" title="%s"></img>' % (
                        image_to_data(images[index]),
                        label,
                        label,
                    )
                    table += "</td>"
                else:
                    table += "<td></td>"
                index += 1
            table += "</tr>"
        table += "</table>"
        output = HTML(table)

    if return_type == "display":
        if clear:
            if clear_output is not None:
                clear_output(wait=True)

        display(output)
    else:
        return output


def progress_bar(range, show_progress=True, progress_type="tqdm"):
    """
    Wrap a range/iter in a progress bar (or not).
    """
    try:
        import tqdm
        import tqdm.notebook
    except ImportError:
        tqdm = None

    if progress_type is None or tqdm is None or show_progress is False:
        return range
    elif progress_type == "tqdm":
        return tqdm.tqdm(range)
    elif (progress_type == "notebook") and (sys.platform != "emscripten"):
        return tqdm.notebook.tqdm(range)
    else:
        return range


def images_to_movie(
    *frames, movie_name="aitk_movie", start=0, stop=None, loop=0, duration=100, mp4=True
):
    """
    Save as animated gif and optionally mp4; show with controls.
    loop - 0 means continually
    duration - in MS
    """
    if stop is None:
        stop = len(frames) * 0.1

    stop = min(stop, len(frames) * 0.1)

    # First, save animated gif:
    frames[0].save(
        movie_name + ".gif",
        save_all=True,
        append_images=frames[1:],
        loop=loop,
        duration=duration,
    )
    if mp4:
        retval = os.system(
            """ffmpeg -y -v quiet -nostats -hide_banner -loglevel error -i {0}.gif -movflags faststart -pix_fmt yuv420p -vf "scale=trunc(iw/2)*2:trunc(ih/2)*2" {0}.mp4""".format(
                movie_name
            )
        )
        if retval != 0:
            print(
                "created animated gif, but error running ffmpeg; see console log message or use mp4=False in the future"
            )



# -*- coding: utf-8 -*-
# ************************************************************
# aitk.robots: Python robot simulator
#
# Copyright (c) 2021 AITK Developers
#
# https://github.com/ArtificialIntelligenceToolkit/aitk.robots
# ************************************************************

import io
import math

from PIL import Image
from svgwrite import Drawing

from .base import Backend

from aitk.utils import Color


# First attempt, replicating the low-level SVG badly
# Second attempt should implement the HIGH-LEVEL API


class SVGBackend(Backend):
    # Specific to this class:

    def initialize(self):
        self.widget = None
        self.char_width = 6.8
        self.char_height = 11
        self.reset()

    def reset(self):
        self.stack = []
        self.points = []
        dwg = Drawing("canvas.svg", (self.width, self.height))
        dwg.viewbox(0, 0, self.width, self.height)
        self.stack.append(dwg)

    # Overrides:

    def watch(self):
        from ipywidgets import HTML

        if self.widget is None:
            self.widget = HTML(value=self.stack[0]._repr_svg_(),)

        return self.widget

    def draw_watcher(self):
        if self.widget:
            self.widget.value = self.stack[0]._repr_svg_()

    def update_dimensions(self, width, height, scale):
        # No need, SVG handles this
        pass

    def flush(self):
        pass

    def get_image(self, time):
        try:
            import cairosvg
        except ImportError:
            print("This backend.get_image() requires cairosvg")
            return

        bytes = cairosvg.svg2png(self.stack[0].tostring(), scale=self._scale)
        fp = io.BytesIO(bytes)
        picture = Image.open(fp)
        if picture.mode == "RGB":
            picture = picture.convert("RGBA")
        return picture

    # Low-level API:

    def set_stroke_style(self, color):
        self.stroke_style = color
        self.stroke_style_color = Color(0, 0, 0, 255)

    def set_fill_style(self, color):
        self.fill_style = color.rgb()
        self.fill_style_color = color

    def fill_opacity(self):
        alpha = self.fill_style_color.alpha
        if alpha != 255:
            return round(alpha / 255, 2)

    def stroke_opacity(self):
        alpha = self.stroke_style_color.alpha
        if alpha != 255:
            return round(alpha / 255, 2)

    def get_style(self, *items):
        fo = self.fill_opacity()
        so = self.stroke_opacity()
        map = {
            "fill": ("fill:%s" % self.fill_style) if self.fill_style else None,
            "stroke-width": "stroke-width:%s" % self.line_width,
            "stroke": ("stroke:%s" % self.stroke_style) if self.stroke_style else None,
            "fill-opacity": ("fill-opacity:%s" % fo) if fo is not None else None,
            "stroke-opacity": ("stroke-opacity:%s" % so) if so is not None else None,
        }
        styles = []
        for item in items:
            v = map[item]
            if v:
                styles.append(v)
        style = ";".join(styles)
        return style

    def arc(self, x, y, radius, startAngle, endAngle):
        def polarToCartesian(centerX, centerY, radius, angleInRadians):
            return [
                round(centerX + (radius * math.cos(angleInRadians)), 2),
                round(centerY + (radius * math.sin(angleInRadians)), 2),
            ]

        start = polarToCartesian(x, y, radius, endAngle)
        end = polarToCartesian(x, y, radius, startAngle)

        if len(self.points) > 0:
            path = [
                "M",
                start[0],
                start[1],
                "A",
                round(radius, 2),
                round(radius, 2),
                0,
                0,
                0,
                end[0],
                end[1],
                "L",
                self.points[-1][0],
                self.points[-1][1],
            ]
            styles = ["fill", "fill-opacity"]
        else:
            path = [
                "M",
                start[0],
                start[1],
                "A",
                round(radius, 2),
                round(radius, 2),
                0,
                0,
                0,
                end[0],
                end[1],
            ]
            styles = ["stroke", "stroke-width"]
        path = [str(item) for item in path]
        d = " ".join(path)
        dwg = self.stack[-1]
        style = self.get_style(*styles)
        dwg.add(self.stack[0].path(d=d, style=style))

    def clear_rect(self, x, y, width, height):
        self.reset()

    def text(self, text, x, y):
        self.fill_text(text, x, y + self.char_height)

    def fill_text(self, text, x, y):
        style = self.get_style("fill")
        dwg = self.stack[-1]
        dwg.add(self.stack[0].text(text=text, insert=(x, y), style=style))

    def fill_rect(self, x, y, width, height):
        style = self.get_style("fill")
        self.stack[-1].add(
            self.stack[0].rect(
                insert=(round(x, 2), round(y, 2)),
                size=(round(width, 2), round(height, 2)),
                style=style,
            )
        )

    def fill(self):
        if len(self.points) >= 2:
            style = self.get_style("fill")
            dwg = self.stack[-1]
            dwg.add(self.stack[0].polygon(points=self.points, style=style))
            self.points.clear()

    def stroke(self):
        if len(self.points) >= 2:
            style = self.get_style("stroke", "stroke-width", "stroke-opacity")
            dwg = self.stack[-1]
            last = self.points[0]
            for point in self.points[1:]:
                dwg.add(self.stack[0].line(start=last, end=point, style=style))
                last = point
            self.points.clear()

    def move_to(self, x, y):
        self.points = [(round(x, 2), round(y, 2))]

    def line_to(self, x, y):
        self.points.append([round(x, 2), round(y, 2)])

    def save(self):
        dwg = self.stack[0]
        self.stack.append(dwg.add(dwg.g()))

    def restore(self):
        parent = self.stack[-2]
        dwg = self.stack[-1]
        parent.add(dwg)
        self.stack.pop()

    def translate(self, x, y):
        self.stack[-1].translate(x, y)

    def scale(self, xscale, yscale):
        pass

    def set_transform(self, x, y, z, a, b, c):
        pass

    def rotate(self, angle):
        # comes in radians, need to convert to degrees
        self.stack[-1].rotate(angle * 180 / math.pi)

    def begin_path(self):
        pass

    def close_path(self):
        pass

    def ellipse(self, x, y, radiusX, radiusY, a, b, angle):
        dwg = self.stack[-1]
        style = self.get_style("fill")
        dwg.add(
            self.stack[0].ellipse(
                center=(round(x, 2), round(y, 2)),
                r=(round(radiusX, 2), round(radiusY, 2)),
                style=style,
            )
        )

    def get_image_data(self):
        pass

    def put_image_data(self, scaled, x, y):
        pass

    def create_image_data(sefl, width, height):
        pass



"""A NEAT (NeuroEvolution of Augmenting Topologies) implementation"""

# Either do this, or add them in setup.py:
import aitk.algorithms.neat.nn as nn
import aitk.algorithms.neat.ctrnn as ctrnn
import aitk.algorithms.neat.iznn as iznn
import aitk.algorithms.neat.distributed as distributed
import aitk.algorithms.neat.visualize as visualize

from .config import Config
from .population import Population, CompleteExtinctionException
from .genome import DefaultGenome
from .reproduction import DefaultReproduction
from .stagnation import DefaultStagnation
from .reporting import StdOutReporter
from .species import DefaultSpeciesSet
from .statistics import StatisticsReporter
from .parallel import ParallelEvaluator
from .distributed import DistributedEvaluator, host_is_local
from .threaded import ThreadedEvaluator
from .checkpoint import Checkpointer



# -*- coding: utf-8 -*-
# ***********************************************************
# aitk.utils: Python AI utils
#
# Copyright (c) 2020 AITK Developers
#
# https://github.com/ArtificialIntelligenceToolkit/aitk.utils
#
# ***********************************************************

import numpy as np

from .utils import get_file, Dataset


def load_data(data_file):
    np_data = np.load(data_file, allow_pickle=True)

    labels = np_data["labels"]

    return Dataset(
        train_inputs=np_data["data"].tolist(),
        train_features=[labels.split("_") for labels in labels.tolist()]
    )

def get_full():
    """
    624, 120, 128
    """
    data_file = get_file(
        "cmu_faces_full_size.npz",
        "https://raw.githubusercontent.com/ArtificialIntelligenceToolkit/datasets/master/cmu_faces/cmu_faces_full_size.npz",
        extract=True,
    )
    return load_data(data_file)

def get_half():
    """
    """
    data_file = get_file(
        "cmu_faces_half_size.npz",
        "https://raw.githubusercontent.com/ArtificialIntelligenceToolkit/datasets/master/cmu_faces/cmu_faces_half_size.npz",
        extract=True,
    )
    return load_data(data_file)

def get_quarter():
    """
    """
    data_file = get_file(
        "cmu_faces_quarter_size.npz",
        "https://raw.githubusercontent.com/ArtificialIntelligenceToolkit/datasets/master/cmu_faces/cmu_faces_quarter_size.npz",
        extract=True,
    )
    return load_data(data_file)




# -*- coding: utf-8 -*-
# ***********************************************************
# aitk.utils: Python utils for AI
#
# Copyright (c) 2021 AITK Developers
#
# https://github.com/ArtificialIntelligenceToolkit/aitk.utils
#
# ***********************************************************

try:
    from ipywidgets import GridspecLayout, Button, Layout, VBox
except ImportError:
    pass

import math

def control(button):
    print(button.description)

class JoyPad():
    """
    A dynamic joypad made up of buttons.

    Args:
        scale: the scaling of translate and rotate values; defaults (1.0, 1.0)
        width: width of the widget; default 250
        height: height of the widget; default 250
        function: the function to call when clicking a button; default print
    """
    def __init__(self, scale=(1.0, 1.0), width=250, height=250, function=print):
        self.width = "%spx" % width if width is not None else None
        self.height = "%spx" % height if height is not None else None
        self.cell_width = math.floor(width / 3)
        self.cell_height = math.floor(height / 3)
        self.function = function
        self.scale = scale
        self.arrows = [
            "⬉⬆⬈",
            "⬅⊙➡",
            "⬋⬇⬊",
        ]
        self.movement = [
            [(1.0, -1.), (1.0, 0.0), (1.0, 1.0)],
            [(0.0, -1.), (0.0, 0.0), (0.0, 1.0)],
            [(-1., -1.), (-1., 0.0), (-1., 1.0)],
        ]
        self.last_movement = (0, 0)
        self.grid = GridspecLayout(3, 3, width=self.width, height=self.height)
        for row in range(3):
            for col in range(3):
                layout = Layout(
                    width="%spx" % self.cell_width,
                    height="%spx" % self.cell_height,
                    max_width="%spx" % self.cell_width,
                    max_height="%spx" % self.cell_height,
                    margin="0px",
                )
                style = {
                    "font_size": "xx-large",
                }
                self.grid[row, col] = Button(description=self.arrows[row][col], layout=layout, style=style)
                self.grid[row, col].on_click(lambda button, row=row, col=col: self.control(row, col))

        self.continue_button = Button(description="Continue", layout={'width': self.width, 'height': "50px", "padding": "5px"}, style=style)
        self.continue_button.on_click(self.continue_motion)
        self.widget = VBox(
            children=[
                self.grid,
                self.continue_button,
            ]
        )
                                     
    def continue_motion(self, button):
        translate, rotate = self.last_movement
        self.function(
            translate * self.scale[0],
            rotate * self.scale[1],
        )

    def control(self, row, col):
        self.last_movement = self.movement[row][col]
        translate, rotate = self.last_movement
        self.function(
            translate * self.scale[0],
            rotate * self.scale[1],
        )

    def watch(self):
        display(self.widget)

    def get_widget(self):
        return self.widget



# -*- coding: utf-8 -*-
# *******************************************************
# aitk: Python tools for AI
#
# Copyright (c) 2021 AITK Developers
#
# https://github.com/ArtificialIntelligenceToolkit/aitk/
#
# *******************************************************

"""
aitk setup
"""
import io
import os
import json

import setuptools

name = "aitk"
HERE = os.path.abspath(os.path.dirname(__file__))

# Get our version
def get_version(file, name="__version__"):
    """Get the version of the package from the given file by
    executing it and extracting the given `name`.
    """
    path = os.path.realpath(file)
    version_ns = {}
    with io.open(path, encoding="utf8") as f:
        exec(f.read(), {}, version_ns)
    return version_ns[name]

with open(os.path.join(HERE, "README.md"), "r") as fh:
    long_description = fh.read()

version = get_version(os.path.join(HERE, "aitk/_version.py"))

setup_args = dict(
    name=name,
    version=version,
    url="https://github.com/ArtificialIntelligenceToolkit/%s" % name,
    author="Douglas Blank",
    description="Python tools for AI",
    long_description=long_description,
    long_description_content_type="text/markdown",
    package_data={
        "aitk.utils": ["fonts/*.ttf"],
        "aitk.robots": ["worlds/*.json", "worlds/*.png"],
    },
    install_requires=["Pillow", "ipywidgets", "tqdm", "numpy", "matplotlib", "tensorflow>=2.17.0"],
    packages=setuptools.find_packages(),
    python_requires=">=3.9",
    license="BSD-3-Clause",
    platforms="Linux, Mac OS X, Windows",
    keywords=["ai", "artificial intelligence", "robots",
              "simulator", "jupyter", "python", "machine learning",
              "neural networks", "keras", "tensorflow"],
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Framework :: Jupyter",
    ],
)

if __name__ == "__main__":
    setuptools.setup(**setup_args)



from ..graphs import feed_forward_layers


class FeedForwardNetwork(object):
    def __init__(self, inputs, outputs, node_evals):
        self.input_nodes = inputs
        self.output_nodes = outputs
        self.node_evals = node_evals
        self.values = dict((key, 0.0) for key in inputs + outputs)

    def activate(self, inputs):
        if len(self.input_nodes) != len(inputs):
            raise RuntimeError("Expected {0:n} inputs, got {1:n}".format(len(self.input_nodes), len(inputs)))

        for k, v in zip(self.input_nodes, inputs):
            self.values[k] = v

        for node, act_func, agg_func, bias, response, links in self.node_evals:
            node_inputs = []
            for i, w in links:
                node_inputs.append(self.values[i] * w)
            s = agg_func(node_inputs)
            self.values[node] = act_func(bias + response * s)

        return [self.values[i] for i in self.output_nodes]

    @staticmethod
    def create(genome, config):
        """ Receives a genome and returns its phenotype (a FeedForwardNetwork). """

        # Gather expressed connections.
        connections = [cg.key for cg in genome.connections.values() if cg.enabled]

        layers = feed_forward_layers(config.genome_config.input_keys, config.genome_config.output_keys, connections)
        node_evals = []
        for layer in layers:
            for node in layer:
                inputs = []
                node_expr = [] # currently unused
                for conn_key in connections:
                    inode, onode = conn_key
                    if onode == node:
                        cg = genome.connections[conn_key]
                        inputs.append((inode, cg.weight))
                        node_expr.append("v[{}] * {:.7e}".format(inode, cg.weight))

                ng = genome.nodes[node]
                aggregation_function = config.genome_config.aggregation_function_defs.get(ng.aggregation)
                activation_function = config.genome_config.activation_defs.get(ng.activation)
                node_evals.append((node, activation_function, aggregation_function, ng.bias, ng.response, inputs))

        return FeedForwardNetwork(config.genome_config.input_keys, config.genome_config.output_keys, node_evals)



# -*- coding: utf-8 -*-
# ******************************************************
# aitk.networks: Keras model wrapper with visualizations
#
# Copyright (c) 2021 Douglas S. Blank
#
# https://github.com/ArtificialIntelligenceToolkit/aitk.networks
#
# ******************************************************

from tensorflow.keras.layers import *

from .network import Network, SequentialNetwork, SimpleNetwork  # noqa: F401



"""
This module implements a spiking neural network.
Neurons are based on the model described by:

Izhikevich, E. M.
Simple Model of Spiking Neurons
IEEE TRANSACTIONS ON NEURAL NETWORKS, VOL. 14, NO. 6, NOVEMBER 2003

http://www.izhikevich.org/publications/spikes.pdf
"""

from ..attributes import FloatAttribute
from ..genes import BaseGene, DefaultConnectionGene
from ..genome import DefaultGenomeConfig, DefaultGenome
from ..graphs import required_for_output

# a, b, c, d are the parameters of the Izhikevich model.
# a: the time scale of the recovery variable
# b: the sensitivity of the recovery variable
# c: the after-spike reset value of the membrane potential
# d: after-spike reset of the recovery variable
# The following parameter sets produce some known spiking behaviors:
# pylint: disable=bad-whitespace
REGULAR_SPIKING_PARAMS        = {'a': 0.02, 'b': 0.20, 'c': -65.0, 'd': 8.00}
INTRINSICALLY_BURSTING_PARAMS = {'a': 0.02, 'b': 0.20, 'c': -55.0, 'd': 4.00}
CHATTERING_PARAMS             = {'a': 0.02, 'b': 0.20, 'c': -50.0, 'd': 2.00}
FAST_SPIKING_PARAMS           = {'a': 0.10, 'b': 0.20, 'c': -65.0, 'd': 2.00}
THALAMO_CORTICAL_PARAMS       = {'a': 0.02, 'b': 0.25, 'c': -65.0, 'd': 0.05}
RESONATOR_PARAMS              = {'a': 0.10, 'b': 0.25, 'c': -65.0, 'd': 2.00}
LOW_THRESHOLD_SPIKING_PARAMS  = {'a': 0.02, 'b': 0.25, 'c': -65.0, 'd': 2.00}


# TODO: Add mechanisms analogous to axon & dendrite propagation delay.


class IZNodeGene(BaseGene):
    """Contains attributes for the iznn node genes and determines genomic distances."""

    _gene_attributes = [FloatAttribute('bias'),
                        FloatAttribute('a'),
                        FloatAttribute('b'),
                        FloatAttribute('c'),
                        FloatAttribute('d')]

    def distance(self, other, config):
        s = abs(self.a - other.a) + abs(self.b - other.b) \
            + abs(self.c - other.c) + abs(self.d - other.d)
        return s * config.compatibility_weight_coefficient


class IZGenome(DefaultGenome):
    @classmethod
    def parse_config(cls, param_dict):
        param_dict['node_gene_type'] = IZNodeGene
        param_dict['connection_gene_type'] = DefaultConnectionGene
        return DefaultGenomeConfig(param_dict)


class IZNeuron(object):
    """Sets up and simulates the iznn nodes (neurons)."""
    def __init__(self, bias, a, b, c, d, inputs):
        """
        a, b, c, d are the parameters of the Izhikevich model.

        :param float bias: The bias of the neuron.
        :param float a: The time-scale of the recovery variable.
        :param float b: The sensitivity of the recovery variable.
        :param float c: The after-spike reset value of the membrane potential.
        :param float d: The after-spike reset value of the recovery variable.
        :param inputs: A list of (input key, weight) pairs for incoming connections.
        :type inputs: list(tuple(int, float))
        """
        self.a = a
        self.b = b
        self.c = c
        self.d = d
        self.bias = bias
        self.inputs = inputs

        # Membrane potential (millivolts).
        self.v = self.c

        # Membrane recovery variable.
        self.u = self.b * self.v

        self.fired = 0.0
        self.current = self.bias

    def advance(self, dt_msec):
        """
        Advances simulation time by the given time step in milliseconds.

        v' = 0.04 * v^2 + 5v + 140 - u + I
        u' = a * (b * v - u)

        if v >= 30 then
            v <- c, u <- u + d
        """
        # TODO: Make the time step adjustable, and choose an appropriate
        # numerical integration method to maintain stability.
        # TODO: The need to catch overflows indicates that the current method is
        # not stable for all possible network configurations and states.
        try:
            self.v += 0.5 * dt_msec * (0.04 * self.v ** 2 + 5 * self.v + 140 - self.u + self.current)
            self.v += 0.5 * dt_msec * (0.04 * self.v ** 2 + 5 * self.v + 140 - self.u + self.current)
            self.u += dt_msec * self.a * (self.b * self.v - self.u)
        except OverflowError:
            # Reset without producing a spike.
            self.v = self.c
            self.u = self.b * self.v

        self.fired = 0.0
        if self.v > 30.0:
            # Output spike and reset.
            self.fired = 1.0
            self.v = self.c
            self.u += self.d

    def reset(self):
        """Resets all state variables."""
        self.v = self.c
        self.u = self.b * self.v
        self.fired = 0.0
        self.current = self.bias


class IZNN(object):
    """Basic iznn network object."""
    def __init__(self, neurons, inputs, outputs):
        self.neurons = neurons
        self.inputs = inputs
        self.outputs = outputs
        self.input_values = {}

    def set_inputs(self, inputs):
        """Assign input voltages."""
        if len(inputs) != len(self.inputs):
            raise RuntimeError(
                "Number of inputs {0:d} does not match number of input nodes {1:d}".format(
                    len(inputs), len(self.inputs)))
        for i, v in zip(self.inputs, inputs):
            self.input_values[i] = v

    def reset(self):
        """Reset all neurons to their default state."""
        for n in self.neurons.values():
            n.reset()

    def get_time_step_msec(self):
        # pylint: disable=no-self-use
        # TODO: Investigate performance or numerical stability issues that may
        # result from using this hard-coded time step.
        return 0.05

    def advance(self, dt_msec):
        for n in self.neurons.values():
            n.current = n.bias
            for i, w in n.inputs:
                ineuron = self.neurons.get(i)
                if ineuron is not None:
                    ivalue = ineuron.fired
                else:
                    ivalue = self.input_values[i]

                n.current += ivalue * w

        for n in self.neurons.values():
            n.advance(dt_msec)

        return [self.neurons[i].fired for i in self.outputs]

    @staticmethod
    def create(genome, config):
        """ Receives a genome and returns its phenotype (a neural network). """
        genome_config = config.genome_config
        required = required_for_output(genome_config.input_keys, genome_config.output_keys, genome.connections)

        # Gather inputs and expressed connections.
        node_inputs = {}
        for cg in genome.connections.values():
            if not cg.enabled:
                continue

            i, o = cg.key
            if o not in required and i not in required:
                continue

            if o not in node_inputs:
                node_inputs[o] = [(i, cg.weight)]
            else:
                node_inputs[o].append((i, cg.weight))

        neurons = {}
        for node_key in required:
            ng = genome.nodes[node_key]
            inputs = node_inputs.get(node_key, [])
            neurons[node_key] = IZNeuron(ng.bias, ng.a, ng.b, ng.c, ng.d, inputs)

        genome_config = config.genome_config
        return IZNN(neurons, genome_config.input_keys, genome_config.output_keys)



"""
Implementation of reporter classes, which are triggered on particular events. Reporters
are generally intended to  provide information to the user, store checkpoints, etc.
"""
from __future__ import division, print_function

import time

from neat.math_util import mean, stdev


class ReporterSet(object):
    """
    Keeps track of the set of reporters
    and gives methods to dispatch them at appropriate points.
    """
    def __init__(self):
        self.reporters = []

    def add(self, reporter):
        self.reporters.append(reporter)

    def remove(self, reporter):
        self.reporters.remove(reporter)

    def start_generation(self, gen):
        for r in self.reporters:
            r.start_generation(gen)

    def end_generation(self, config, population, species_set):
        for r in self.reporters:
            r.end_generation(config, population, species_set)

    def post_evaluate(self, config, population, species, best_genome):
        for r in self.reporters:
            r.post_evaluate(config, population, species, best_genome)

    def post_reproduction(self, config, population, species):
        for r in self.reporters:
            r.post_reproduction(config, population, species)

    def complete_extinction(self):
        for r in self.reporters:
            r.complete_extinction()

    def found_solution(self, config, generation, best):
        for r in self.reporters:
            r.found_solution(config, generation, best)

    def species_stagnant(self, sid, species):
        for r in self.reporters:
            r.species_stagnant(sid, species)

    def info(self, msg):
        for r in self.reporters:
            r.info(msg)


class BaseReporter(object):
    """Definition of the reporter interface expected by ReporterSet."""
    def start_generation(self, generation):
        pass

    def end_generation(self, config, population, species_set):
        pass

    def post_evaluate(self, config, population, species, best_genome):
        pass

    def post_reproduction(self, config, population, species):
        pass

    def complete_extinction(self):
        pass

    def found_solution(self, config, generation, best):
        pass

    def species_stagnant(self, sid, species):
        pass

    def info(self, msg):
        pass


class StdOutReporter(BaseReporter):
    """Uses `print` to output information about the run; an example reporter class."""
    def __init__(self, show_species_detail):
        self.show_species_detail = show_species_detail
        self.generation = None
        self.generation_start_time = None
        self.generation_times = []
        self.num_extinctions = 0

    def start_generation(self, generation):
        self.generation = generation
        print('\n ****** Running generation {0} ****** \n'.format(generation))
        self.generation_start_time = time.time()

    def end_generation(self, config, population, species_set):
        ng = len(population)
        ns = len(species_set.species)
        if self.show_species_detail:
            print('Population of {0:d} members in {1:d} species:'.format(ng, ns))
            print("   ID   age  size  fitness  adj fit  stag")
            print("  ====  ===  ====  =======  =======  ====")
            for sid in sorted(species_set.species):
                s = species_set.species[sid]
                a = self.generation - s.created
                n = len(s.members)
                f = "--" if s.fitness is None else "{:.1f}".format(s.fitness)
                af = "--" if s.adjusted_fitness is None else "{:.3f}".format(s.adjusted_fitness)
                st = self.generation - s.last_improved
                print(
                    "  {: >4}  {: >3}  {: >4}  {: >7}  {: >7}  {: >4}".format(sid, a, n, f, af, st))
        else:
            print('Population of {0:d} members in {1:d} species'.format(ng, ns))

        elapsed = time.time() - self.generation_start_time
        self.generation_times.append(elapsed)
        self.generation_times = self.generation_times[-10:]
        average = sum(self.generation_times) / len(self.generation_times)
        print('Total extinctions: {0:d}'.format(self.num_extinctions))
        if len(self.generation_times) > 1:
            print("Generation time: {0:.3f} sec ({1:.3f} average)".format(elapsed, average))
        else:
            print("Generation time: {0:.3f} sec".format(elapsed))

    def post_evaluate(self, config, population, species, best_genome):
        # pylint: disable=no-self-use
        fitnesses = [c.fitness for c in population.values()]
        fit_mean = mean(fitnesses)
        fit_std = stdev(fitnesses)
        best_species_id = species.get_species_id(best_genome.key)
        print('Population\'s average fitness: {0:3.5f} stdev: {1:3.5f}'.format(fit_mean, fit_std))
        print(
            'Best fitness: {0:3.5f} - size: {1!r} - species {2} - id {3}'.format(best_genome.fitness,
                                                                                 best_genome.size(),
                                                                                 best_species_id,
                                                                                 best_genome.key))

    def complete_extinction(self):
        self.num_extinctions += 1
        print('All species extinct.')

    def found_solution(self, config, generation, best):
        print('\nBest individual in generation {0} meets fitness threshold - complexity: {1!r}'.format(
            self.generation, best.size()))

    def species_stagnant(self, sid, species):
        if self.show_species_detail:
            print("\nSpecies {0} with {1} members is stagnated: removing it".format(sid, len(species.members)))

    def info(self, msg):
        print(msg)



# -*- coding: utf-8 -*-
# *************************************
# aitk.robots: Python robot simulator
#
# Copyright (c) 2020 Calysto Developers
#
# https://github.com/ArtificialIntelligenceToolkit/aitk.robots
#
# *************************************

from .config import setup_backend, switch_backend  # noqa: F401
from .devices import Camera, GroundCamera, LightSensor, RangeSensor, SmellSensor, Compass  # noqa: F401
from .robot import Robot, Scribbler, Vehicle  # noqa: F401
from .utils import gallery, load_world  # noqa: F401
from .world import Beacon, Bulb, Wall, World  # noqa: F401

setup_backend()  # checks os.environ

import time

# Test to see if time.sleep() works:
start_time = time.monotonic()
time.sleep(0.1)
if time.monotonic() - start_time < 0.1:
    print("time.sleep() failed; redefined")
    # It didn't work, we replace time.sleep:
    # happens with current Pyodide
    def sleep(seconds):
        start_time = time.monotonic()
        while time.monotonic() - start_time < seconds:
            pass
    time.sleep = sleep



"""Commonly used functions not available in the Python2 standard library."""
from __future__ import division

from math import sqrt, exp


def mean(values):
    values = list(values)
    return sum(map(float, values)) / len(values)


def median(values):
    values = list(values)
    values.sort()
    return values[len(values) // 2]


def median2(values):
    values = list(values)
    n = len(values)
    if n <= 2:
        return mean(values)
    values.sort()
    if (n % 2) == 1:
        return values[n//2]
    i = n//2
    return (values[i - 1] + values[i])/2.0


def variance(values):
    values = list(values)
    m = mean(values)
    return sum((v - m) ** 2 for v in values) / len(values)


def stdev(values):
    return sqrt(variance(values))


def softmax(values):
    """
    Compute the softmax of the given value set, v_i = exp(v_i) / s,
    where s = sum(exp(v_0), exp(v_1), ..)."""
    e_values = list(map(exp, values))
    s = sum(e_values)
    inv_s = 1.0 / s
    return [ev * inv_s for ev in e_values]


# Lookup table for commonly used {value} -> value functions.
stat_functions = {'min': min, 'max': max, 'mean': mean, 'median': median,
                  'median2': median2}



# -*- coding: utf-8 -*-
# *************************************
# aitk.robots: Python robot simulator
#
# Copyright (c) 2020 Calysto Developers
#
# https://github.com/ArtificialIntelligenceToolkit/aitk.robots
#
# *************************************

import os

from PIL import Image

from .utils import get_file, round_to_nearest

_filename = get_file(
    "coil-100-no-background.zip",
    "https://media.githubusercontent.com/media/Calysto/conx-data/master/coil/coil-100-no-background.zip",
    extract=True,
)

DATA_DIR = os.path.splitext(_filename)[0]


def get_range():
    # start, stop, increment
    return {
        "object-start": 1,
        "object-stop": 100,
        "slice-start": 0,
        "slice-stop": 355,
        "slice-increment": 5,
    }


def get(obj_num, degree):
    slice = degree % 360  # map into range 0 to 360
    slice = min(max(round_to_nearest(slice, 5), 0), 355)  # get the nearest 5th degree
    filename = os.path.join(DATA_DIR, "obj%s__%s.png" % (obj_num, slice))
    image = Image.open(filename)
    return image
